set scriptdir [file dir [info script]]
set root [file normalize [file join $scriptdir ..]]

set fp [open [file join $root configure.ac] r]
set str [read $fp]
close $fp

regexp -- {m4_define\(\[plugin_major_version\], +\[(\d+)\]\)} $str ma m1
regexp -- {m4_define\(\[plugin_minor_version\], +\[(\d+)\]\)} $str ma m2
regexp -- {m4_define\(\[plugin_micro_version\], +\[(\d+)\]\)} $str ma m3

set version ${m1}.${m2}.${m3}

regexp -- {m4_define\(\[plugin_name\], +\[([^ ]+?)\]\)} $str ma m4

set name ${m4}

proc win {} {
	set win32dir ${::name}-${::version}-win32-bin
	set win64dir ${::name}-${::version}-win64-bin

	foreach dst [list $win32dir $win64dir] src [list Win32 x64] {
	    if {[file exists $dst]} {
		file delete -force $dst
	    }
	    file mkdir $dst
	    file copy [file join $::root win $src Release ${::name}.exe] $dst
	    file copy [file join $::scriptdir README.txt] $dst

	    exec 7za a ${dst}.7z $dst
	    file delete -force $dst
	}
}

proc osx {} {
	set src [file join ${::root} src] 
	set dst ${::name}-${::version}-osx-bin
	if {[file exists $dst]} {
		file delete -force $dst
	}
	file mkdir $dst
	file copy [file join $src $::name] $dst
	file copy [file join $::scriptdir README.txt] $dst

	exec zip -ur ${dst}.zip $dst
	file delete -force $dst
}

proc linux {} {
	set src [file join ${::root} src] 
	set dst ${::name}-${::version}-linux-$::tcl_platform(machine)-bin
	if {[file exists $dst]} {
		file delete -force $dst
	}
	file mkdir $dst
	file copy [file join $src $::name] $dst
	exec strip [file join $dst $::name]
	file copy [file join $::scriptdir README.txt] $dst

	exec tar cJvf ${dst}.tar.xz $dst
	file delete -force $dst
}


if {$tcl_platform(platform) eq "windows"} { win; exit; }
if {$tcl_platform(os) eq "Darwin"} { osx; exit; }
if {$tcl_platform(os) eq "Linux"} {linux; exit;}
puts "ERROR! UNKNOWN PLATFORM!!"


